package com.uoc.tellmetweet.bo;

import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import com.uoc.tellmetweet.common.TellmetweetConstants;
import com.uoc.tellmetweet.dao.PaymentsDAO;
import com.uoc.tellmetweet.dao.UserDAO;
import com.uoc.tellmetweet.dao.UserLocationDAO;
import com.uoc.tellmetweet.domain.TmLanguages;
import com.uoc.tellmetweet.domain.TmPayments;
import com.uoc.tellmetweet.domain.TmRols;
import com.uoc.tellmetweet.domain.TmUserlocations;
import com.uoc.tellmetweet.domain.TmUsers;
import com.uoc.tellmetweet.dto.UserDTO;
import com.uoc.tellmetweet.exception.TellmetweetBOException;
import com.uoc.tellmetweet.utils.TellmetweetUtils;

@Component
@Transactional
public class UserBO {
	
	private static Logger logger = LoggerFactory.getLogger(UserBO.class);

	@Autowired
	UserDAO userDAO;
	
	@Autowired
	PaymentsDAO paymentsDAO;
	
	@Autowired
	UserLocationDAO userLocationDAO;
	
	/**
	 * Obtiene el usuario logueado
	 * @param filterTrendDTO
	 * @return List<ChartDTO>
	 * @throws TellmetweetBOException
	 */
	public UserDTO loginUser(String email, String pass) throws TellmetweetBOException {
		
		UserDTO userDTO = null;
		try {
			TmUsers tmUsers = userDAO.loginUser(email, pass);
			if(tmUsers!=null){
				userDTO = new UserDTO();
				userDTO.setId(tmUsers.getId());
				userDTO.setEmail(tmUsers.getEmail());
				userDTO.setName(tmUsers.getName());
				userDTO.setSurname1(tmUsers.getSurname1());
				userDTO.setSurname2(tmUsers.getSurname2());
				userDTO.setCdrol(tmUsers.getCdrol().getId());
			} 

		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new TellmetweetBOException(e.getMessage(), e);
		}
		return userDTO;
	}
	/* end findAllChartByWoeid */
	
	
	/**
	 * Crea un nuevo usuario
	 * @param userDTO
	 * @return Integer
	 * @throws TellmetweetBOException
	 */
	public UserDTO registerUser(UserDTO userDTO) throws TellmetweetBOException {
		
		TmUsers tmUsers = new TmUsers();
		try {
			tmUsers.setActive(1);
			tmUsers.setCdlanguage(TmLanguages.findTmLanguages(TellmetweetConstants.LANGUAGE_ES));
			tmUsers.setCdrol(TmRols.findTmRols(TellmetweetConstants.ROL_REGISTERED));
			tmUsers.setCduseract(TmUsers.findTmUsers(1).getCduseract());
			tmUsers.setEmail(userDTO.getEmail());
			tmUsers.setName(userDTO.getName());
			tmUsers.setPass(TellmetweetUtils.passwordToMd5(userDTO.getPass()));
			tmUsers.setSurname1(userDTO.getSurname1());
			tmUsers.setSurname2(userDTO.getSurname2());
			tmUsers.setFact(TellmetweetUtils.dateToCalendar(new Date()));
			userDTO.setPass(null);
			userDAO.registerUser(tmUsers);
			TmUserlocations tmUserlocations = new TmUserlocations();
			tmUserlocations.setCduser(TmUsers.findTmUsers(tmUsers.getId()));
			tmUserlocations.setFupdate(null);
			tmUserlocations.setWoeid1(TellmetweetConstants.woeid1);
			tmUserlocations.setWoeid2(TellmetweetConstants.woeid2);
			tmUserlocations.setWoeid3(TellmetweetConstants.woeid3);
			userLocationDAO.create(tmUserlocations);
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new TellmetweetBOException(e.getMessage(), e);
		}
		return userDTO;
	}
	/* end registerUser */
	
	
	/**
	 * Comprueba si el usuario tiene permisos
	 * @param userEmail
	 * @return
	 * @throws TellmetweetBOException
	 */
	public Boolean hashPermission(String userEmail) throws TellmetweetBOException {
		
		Boolean permission = false;
		try {
			Integer querysUsed = userDAO.querysUsed(userEmail);
			Integer querysAvailable = userDAO.querysAvailableByRol(userEmail);
			if(querysUsed<querysAvailable){
				userDAO.sumOneQuery(userEmail);
				permission=true;
			}
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new TellmetweetBOException(e.getMessage(), e);
		}
		
		return permission;
	}
	
	
	/**
	 * Cambia el rol del usuario a Premium e inserta el pago en la tabla payments
	 * @param userEmail
	 * @return UserDTO
	 * @throws TellmetweetBOException
	 */
	public UserDTO setPremium(String userEmail) throws TellmetweetBOException {
		
		UserDTO userDTO = new UserDTO();
		try {
			TmUsers tmUsers = userDAO.findTmUsersByEmail(userEmail);
			tmUsers.setCdrol(TmRols.findTmRols(TellmetweetConstants.ROL_PREMIUM));
			userDAO.updateUser(tmUsers);
			TmPayments tmPayments = new TmPayments();
			tmPayments.setCduser(userDAO.findTmUsersByEmail(userEmail).getId());
			tmPayments.setQuantity(Integer.parseInt(TellmetweetConstants.CREDIT_CARD_PAYPAL_PAYMENT_TOTAL));
			tmPayments.setFcreate(TellmetweetUtils.dateToCalendar(new Date()));
			tmPayments.setFpay(TellmetweetUtils.dateToCalendar(new Date()));
			paymentsDAO.save(tmPayments);
			userDTO.setCdrol(tmUsers.getCdrol().getId());
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new TellmetweetBOException(e.getMessage(), e);
		}
		
		return userDTO;
	}
	
	
	/**
	 * Comprueba si el usuario existe en el sistema
	 * @param userEmail
	 * @return Boolean
	 * @throws TellmetweetBOException
	 */
	public Boolean existUser(String userEmail) throws TellmetweetBOException {
		
		Boolean exist = false;
		try {
			TmUsers tmUsers = userDAO.findTmUsersByEmail(userEmail);
			if(tmUsers!=null && tmUsers.getId()!=null){
				exist = true;
			}
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new TellmetweetBOException(e.getMessage(), e);
		}
		return exist;
	}
	
	
	
	
}
