package com.uoc.tellmetweet.dao;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;

import com.uoc.tellmetweet.common.GenericDAOImpl;
import com.uoc.tellmetweet.common.TellmetweetConstants;
import com.uoc.tellmetweet.domain.TmLocations;
import com.uoc.tellmetweet.exception.TellmetweetDAOException;

import twitter4j.Location;
import twitter4j.ResponseList;
import twitter4j.Twitter;
import twitter4j.TwitterException;
import twitter4j.TwitterFactory;

@Transactional
@Repository
public class LocationsDAOImpl extends GenericDAOImpl implements LocationsDAO {
	
	private static Logger logger = LoggerFactory.getLogger(LocationsDAOImpl.class);
	
	@Autowired
	TwitterUtilsDAO twitterUtilsDAO;
	
	
	/**
	 * Persiste un objeto Locations
	 */
	public void saveLocation(TmLocations tmLocations) throws TellmetweetDAOException {
		try {
			save(tmLocations);
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
	}
	/* end saveLocation */
	
	
	/**
	 * Obtiene la lista de locations
	 */
	public ResponseList<Location> findAllLocations(TwitterFactory tf) throws TellmetweetDAOException {
		
		logger.info("Obteniendo Locations");
		ResponseList<Location> dailyTrends = null;
		try {
			Twitter twitter = tf.getInstance();
			
			/* Comprobar límite disponible */
			twitterUtilsDAO.checkLimit(tf, TellmetweetConstants.rateLimit_trendsAvailable);
			dailyTrends = twitter.getAvailableTrends();
			
		} catch (TwitterException t) {
			logger.error("Error: " + t.getMessage(), t);
			throw new TellmetweetDAOException(t.getMessage(), t);
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		
		return dailyTrends;
	}
	/* end findAllLocations */
	

	/**
	 * Obtiene un Location pasando el woeid
	 * Si no se ha insertado ningún registro duplicado, la lista solamente tendrá un registro
	 * @param woeid
	 * @return List<Locations>
	 */
	public List<TmLocations> findLocationByWoeid(Integer woeid) throws TellmetweetDAOException {

		List<TmLocations> listTmLocations = new ArrayList<TmLocations>();
		
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select o.id from TmLocations o ");
			sb.append(" where o.woeid=:woeid");
			
			listTmLocations = TmLocations.entityManager()
				.createQuery(sb.toString())
				.setParameter("woeid", woeid)
				.getResultList();
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		
		return listTmLocations;
	}
	/* end findLocationByWoeid */
	
	
	/**
	 * Comprueba si el location existe en la BD, si no existe lo inserta
	 * @param listLocations
	 */
	public void saveAllLocations(ResponseList<Location> listLocations) throws TellmetweetDAOException {
		
		try {
			for (Location location : listLocations) {
				/* Comprobar si ya existe el location en la BD */
				TmLocations loc = new TmLocations();
				
				if(findLocationByWoeid(location.getWoeid()).isEmpty()){
					/* No existe, se inseta en la BD */
					loc.setCountrycode(location.getCountryCode());
					loc.setCountryname(location.getCountryName());
					loc.setFcreate(Calendar.getInstance());
					loc.setName(location.getName());
					loc.setPlacecode(location.getPlaceCode());
					loc.setPlacename(location.getName());
					loc.setUrl(location.getURL());
					loc.setWoeid(location.getWoeid());
					saveLocation(loc);
				}
				
			}
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
	}
	/* end saveAllLocations */
	
}
