package com.uoc.tellmetweet.dao;


import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.TypedQuery;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;

import com.uoc.tellmetweet.common.GenericDAOImpl;
import com.uoc.tellmetweet.common.TellmetweetConstants;
import com.uoc.tellmetweet.domain.TmActivelocations;
import com.uoc.tellmetweet.domain.TmTrend;
import com.uoc.tellmetweet.domain.TmTrends;
import com.uoc.tellmetweet.dto.FilterTrendDTO;
import com.uoc.tellmetweet.exception.TellmetweetDAOException;
import com.uoc.tellmetweet.utils.TellmetweetUtils;

import twitter4j.Trend;
import twitter4j.Twitter;
import twitter4j.TwitterException;
import twitter4j.TwitterFactory;

@Transactional
@Repository
public class TrendDAOImpl extends GenericDAOImpl implements TrendDAO {

	private static Logger logger = LoggerFactory.getLogger(TrendDAOImpl.class);
	
	@Autowired
	TwitterUtilsDAO twitterUtilsDAO;
	

	/**
	 * Obtiene la lista de trend de un location 
	 * @param tf - TwitterFactory
	 * @param woeid
	 * @return Trend[]
	 */
	public Trend[] findAllTrendByLocation(TwitterFactory tf, TmActivelocations tmActivelocations) throws TellmetweetDAOException {
		
		logger.debug("findAllTrendByLocation");
	
	
		Trend[] trend = null;
		try {
			Twitter twitter = tf.getInstance();
			/* Comprobar límite disponible */
			twitterUtilsDAO.checkLimit(tf, TellmetweetConstants.rateLimit_trendsPlace);
			trend = twitter.getPlaceTrends(tmActivelocations.getLocation().getWoeid()).getTrends();
			
		} catch (TwitterException e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		
		return trend;
	}
	/* end findAllTrends */
	
	
	/**
	 * Obtiene los trend filtrado por el woeid
	 * @param woeid
	 * @throws TellmetweetDAOException
	 */
	public List<TmTrend> findAllTrendByWoeid(FilterTrendDTO filterTrendDTO, int page, int limit, int start) throws TellmetweetDAOException {

		List<TmTrend> tmTrendList = new ArrayList<TmTrend>();
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select o from TmTrend o where o.trends.location.woeid=:woeid");
			if(filterTrendDTO.getFrom_date()!=null){
				sb.append(" and DATE(o.fcreate) >= DATE(:from_date)");
			}
			if(filterTrendDTO.getTo_date()!=null){
				sb.append(" and DATE(o.fcreate) <= DATE(:to_date)");
			}
			
			TypedQuery<TmTrend> query = TmTrends.entityManager().createQuery(sb.toString(), TmTrend.class )
			.setParameter("woeid", filterTrendDTO.getWoeid())
			.setFirstResult(start);
			
			if(filterTrendDTO.getFrom_date()!=null){
				query.setParameter("from_date", filterTrendDTO.getFrom_date());
			}
			if(filterTrendDTO.getTo_date()!=null){
				query.setParameter("to_date", filterTrendDTO.getTo_date());
			}
			
			tmTrendList = (List<TmTrend>) query.getResultList();
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		return tmTrendList;
	}
	/* end findAllTrendByWoeid */
	

	/**
	 * Obtiene el número de HomeTrends by woeid
	 * @param filterTrendDTO
	 * @return int
	 */
	public int countFindAllTrendByWoeid(FilterTrendDTO filterTrendDTO) throws TellmetweetDAOException {
		
		int n = 0;
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select count(o) from TmTrend o where o.trends.location.woeid=:woeid");
			if(filterTrendDTO.getFrom_date()!=null){
				sb.append(" and DATE(o.fcreate) >= DATE(:from_date)");
			}
			if(filterTrendDTO.getTo_date()!=null){
				sb.append(" and DATE(o.fcreate) <= DATE(:to_date)");
			}
			
			TypedQuery<TmTrend> query = TmTrends.entityManager().createQuery(sb.toString(), TmTrend.class )
			.setParameter("woeid", filterTrendDTO.getWoeid());
			
			if(filterTrendDTO.getFrom_date()!=null){
				query.setParameter("from_date", filterTrendDTO.getFrom_date());
			}
			if(filterTrendDTO.getTo_date()!=null){
				query.setParameter("to_date", filterTrendDTO.getTo_date());
			}
			
			BigDecimal count = (BigDecimal) TmTrend.entityManager().createNativeQuery(sb.toString())
					.getSingleResult();
			n = count.intValue();
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		return n;
	}
	/* end countFindAllTrendByWoeid */
	
	
}
