package com.uoc.tellmetweet.dao;

import java.util.Date;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;

import com.uoc.tellmetweet.common.GenericDAOImpl;
import com.uoc.tellmetweet.domain.TmQuerys;
import com.uoc.tellmetweet.domain.TmUsers;
import com.uoc.tellmetweet.exception.TellmetweetBOException;
import com.uoc.tellmetweet.exception.TellmetweetDAOException;
import com.uoc.tellmetweet.utils.TellmetweetUtils;

@Transactional
@Repository
public class UserDAOImpl extends GenericDAOImpl implements UserDAO {

	private static Logger logger = LoggerFactory.getLogger(UserDAOImpl.class);
	
    @PersistenceContext
    EntityManager entityManager;
    
    
    /**
     * Obtiene el usuario logueado
     * @param email
     * @param pass
     * @return TmUsers
     * @throws TellmetweetDAOException
     */
    public TmUsers loginUser(String email, String pass) throws TellmetweetDAOException {
    	
    	TmUsers TmUsers = new TmUsers();
    	
    	try {
    		TmUsers = entityManager.createQuery("SELECT o FROM TmUsers o where o.email=:email and o.pass=:pass", TmUsers.class)
            		.setParameter("email", email)
            		.setParameter("pass", TellmetweetUtils.passwordToMd5(pass))
            		.setMaxResults(1)
            		.getSingleResult();
    	} catch (EmptyResultDataAccessException nr) {
    		return null;
    	} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
    	}
        return TmUsers;
    }
    
    
    /**
     * Crea un nuevo usuario
     * @param tmUsers
     * @return Integer
     * @throws TellmetweetDAOException
     */
    public Integer registerUser(TmUsers tmUsers) throws TellmetweetDAOException {
    	
    	int n=0;
    	try {
    		tmUsers.persist();
    		n = 1;
    	} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
    	}
        return n;
    }
    
    
    /**
     * Comprueba las querys realizadas en el día actual
     * @param userEmail
     * @return Boolean
     * @throws TellmetweetBOException
     */
    public Integer querysUsed(String userEmail) throws TellmetweetBOException {
		
    	Integer querys = 0;
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select count(o.id) from TmQuerys o "
					+ " where DATE(o.fcreate)=:date"
					+ " and o.cduser.email=:email");
			querys = (Integer) TmQuerys.entityManager().createQuery(sb.toString(), java.lang.Long.class )
					.setParameter("date", TellmetweetUtils.getCalendarDateWithoutTime(new Date()))
					.setParameter("email", userEmail)
					.getSingleResult().intValue();
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		return querys;
    }
   
    
    /**
     * Suma una query al usuario
     * @param userEmail
     * @throws TellmetweetBOException
     */
    public void sumOneQuery(String userEmail) throws TellmetweetBOException {
		TmQuerys tmQuerys = new TmQuerys();
		tmQuerys.setFcreate(TellmetweetUtils.dateToCalendar(new Date()));
		tmQuerys.setCduser(findTmUsersByEmail(userEmail));
		tmQuerys.persist();
    }
    
    
    /**
     * Obtiene el número de querys dependiendo del rol del usuario
     * @param userEmail
     * @return
     * @throws TellmetweetBOException
     */
    public Integer querysAvailableByRol(String userEmail) throws TellmetweetBOException {
		
    	Integer querys = 0;
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select o.cdrol.availablequerys from TmUsers o"
					+ " where o.email=:email");
			querys = (Integer) TmQuerys.entityManager().createQuery(sb.toString(), java.lang.Integer.class )
					.setParameter("email", userEmail)
					.getSingleResult();
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		return querys;
    }
    
    
   /**
    * Obtiene un TmUsers por email
    * @param email
    * @return
    */
   public TmUsers findTmUsersByEmail(String email) throws TellmetweetBOException {
	   
	   TmUsers tmUsers = new TmUsers();
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("select o from TmUsers o where o.email=:email");
			List<TmUsers> result = (List<TmUsers>) TmUsers.entityManager().createQuery(sb.toString())
					.setParameter("email", email)
					.setMaxResults(1)
					.getResultList();
			
			if(result!=null && result.size()>0){
				tmUsers = result.get(0);
			}
			
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
		return tmUsers;
   }
   
   
	/**
	 * Actualiza un objeto TmUsers en la base de datos
	 */
	public void updateUser(TmUsers tmUsers) throws TellmetweetDAOException {
		
		try {
			tmUsers.merge();
		} catch (Exception e) {
			logger.error("Error: " + e.getMessage(), e);
			throw new TellmetweetDAOException(e.getMessage(), e);
		}
	}
   
}
