/* Copyright (c) 2015 The Open Source Geospatial Foundation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * A store that is synchronized with a GeoExt.data.store.Layers. It can be used
 * by an Ext.tree.Panel.
 *
 * @class GeoExt.data.store.LayersTree
 */
Ext.define('GeoExt.data.store.LayersTree', {
    extend: 'Ext.data.TreeStore',

    alternateClassName: ['GeoExt.data.TreeStore'],

    requires: [
        'GeoExt.util.Layer'
    ],

    mixins: [
        'GeoExt.mixin.SymbolCheck'
    ],

    // <debug>
    symbols: [
        'ol.Collection',
        'ol.Collection#getArray',
        'ol.Collection#once',
        'ol.Collection#un',
        'ol.layer.Base',
        'ol.layer.Base#get',
        'ol.layer.Group',
        'ol.layer.Group#get',
        'ol.layer.Group#getLayers'
    ],
    // </debug>

    model: 'GeoExt.data.model.LayerTreeNode',

    config: {
        /**
         * The ol.layer.Group that the tree is derived from.
         *
         * @cfg {ol.layer.Group}
         */
        layerGroup: null,

        /**
         * The layer property that will be used to label tree nodes.
         *
         * @cfg {String}
         */
        textProperty: 'name',

        /**
         * Configures the behaviour of the checkbox of an ol.layer.Group
         * (folder). Possible values are 'classic' or 'ol3'.
         * 'classic' forwards the checkstate to the children of the folder.
         * 'ol3' emulates the behaviour of ol.layer.Group. So an layerGroup can
         * be invisible but can have visible children.
         * 'classic':
         *   - Check a leaf --> all parent nodes are checked
         *   - Uncheck all leafs in a folder --> parent node is unchecked
         *   - Check a folder Node --> all children are checked
         *   - Uncheck a folder Node --> all children are unchecked
         * 'ol3':
         *   - Emulates the behaviour of an ol.layer.Group, so a parentfolder
         *     can be unchecked but still contain checked leafs and vice versa.
         * @cfg
         */
        folderToggleMode: 'classic'
    },

    /**
     * Defines if the order of the layers added to the store will be
     * reversed. The default behaviour and what most users expect is
     * that mapLayers on top are also on top in the tree.
     *
     * @property {Boolean}
     */
    inverseLayerOrder: true,

    /**
     * Whether the treestore currently shall handle openlayers collection
     * change events. See #suspendCollectionEvents and #resumeCollectionEvents.
     *
     * @property
     * @private
     */
    collectionEventsSuspended: false,

    /**
     * @cfg
     * @inheritdoc Ext.data.TreeStore
     */
    proxy: {
        type: 'memory',
        reader: {
            type: 'json'
        }
    },

    root: {
        expanded: true
    },

    constructor: function(){
        var me = this;
        me.callParent(arguments);

        var collection = me.layerGroup.getLayers();
        Ext.each(collection.getArray(), function(layer) {
            me.addLayerNode(layer);
        }, me, me.inverseLayerOrder);

        me.bindGroupLayerCollectionEvents(me.layerGroup);

        me.on({
            remove: me.handleRemove,
            noderemove: me.handleNodeRemove,
            nodeappend: me.handleNodeAppend,
            nodeinsert: me.handleNodeInsert,
            scope: me
        });
    },

    /**
     * Apllies the folderToggleMode to the treenodes.
     * @private
     */
    applyFolderToggleMode: function(folderToggleMode){
        if(folderToggleMode === 'classic' || folderToggleMode === 'ol3'){
            var rootNode = this.getRootNode();
            if(rootNode){
                rootNode.cascadeBy({
                    before: function(child){
                        child.set('__toggleMode', folderToggleMode);
                    }
                });
            }
        } else {
            Ext.raise("Invalid folderToggleMode set in " + this.self.getName()
                + ": " + folderToggleMode + ". 'classic' or 'ol3' are valid.");
        }
        return folderToggleMode;
    },

    /**
     * Listens to the remove event and syncs the attached layergroup.
     *
     * @param {GeoExt.data.store.LayersTree} store The layer store.
     * @param {GeoExt.data.model.LayerTreeNode[]} records An array of the
     *     removed nodes.
     * @private
     */
    handleRemove: function(store, records){
        var me = this;
        me.suspendCollectionEvents();
        Ext.each(records, function(record) {
            var layerOrGroup = record.getOlLayer();
            if(layerOrGroup instanceof ol.layer.Group){
                me.unbindGroupLayerCollectionEvents(layerOrGroup);
            }
            var group = GeoExt.util.Layer.findParentGroup(
                layerOrGroup, me.getLayerGroup()
            );
            if (!group) {
                group = me.getLayerGroup();
            }
            if (group) {
                group.getLayers().remove(layerOrGroup);
            }
        });
        me.resumeCollectionEvents();
    },

    /**
     * Listens to the noderemove event. Updates the tree with the current
     * map state.
     *
     * @param {GeoExt.data.model.LayerTreeNode} parentNode
     * @param {GeoExt.data.model.LayerTreeNode} removedNode
     * @private
     */
    handleNodeRemove: function(parentNode, removedNode){
        var me = this;
        var layerOrGroup = removedNode.getOlLayer();
        if (!layerOrGroup) {
            layerOrGroup = me.getLayerGroup();
        }
        if(layerOrGroup instanceof ol.layer.Group){
            me.unbindGroupLayerCollectionEvents(layerOrGroup);
        }
        var group = GeoExt.util.Layer.findParentGroup(
            layerOrGroup, me.getLayerGroup()
        );

        if (group) {
            me.suspendCollectionEvents();
            group.getLayers().remove(layerOrGroup);
            me.resumeCollectionEvents();
        }
    },

    handleNodeAppend: function(parentNode, appendedNode){
        var me = this;
        var group = parentNode.getOlLayer();
        var layer = appendedNode.getOlLayer();

        if (!group) {
            group = me.getLayerGroup();
        }

        // check if the layer is possibly already at the desired index:
        var layerInGroupIdx = GeoExt.util.Layer.getLayerIndex(
            layer, group
        );
        if (layerInGroupIdx === -1) {
            me.suspendCollectionEvents();
            if (me.inverseLayerOrder) {
                group.getLayers().insertAt(0, layer);
            } else {
                group.getLayers().push(layer);
            }
            me.resumeCollectionEvents();
        }
    },

    handleNodeInsert: function(parentNode, insertedNode, insertedBefore){
        var me = this;
        var group = parentNode.getOlLayer();
        if (!group) {
            // can only happen if a node was dragged before the visible root.
            group = me.getLayerGroup();
        }
        var layer = insertedNode.getOlLayer();
        var beforeLayer = insertedBefore.getOlLayer();
        var groupLayers = group.getLayers();
        var beforeIdx = GeoExt.util.Layer.getLayerIndex(beforeLayer, group);
        var insertIdx = beforeIdx;
        if (me.inverseLayerOrder) {
            insertIdx += 1;
        }

        // check if the layer is possibly already at the desired index:
        var currentLayerInGroupIdx = GeoExt.util.Layer.getLayerIndex(
            layer, group
        );
        if (currentLayerInGroupIdx !== insertIdx) {
            me.suspendCollectionEvents();
            groupLayers.insertAt(insertIdx, layer);
            me.resumeCollectionEvents();
        }
    },

    /**
     * Adds a layer as a node to the store. It can be an ol.layer.Base.
     *
     * @param {Ext.data.NodeInterface} node
     */
    addLayerNode: function(layerOrGroup){
        var me = this;
        // 2. get group to which the layer was added
        var group = GeoExt.util.Layer.findParentGroup(
            layerOrGroup, me.getLayerGroup()
        );

        // 3. get index of layer in that group
        var layerIdx = GeoExt.util.Layer.getLayerIndex(layerOrGroup, group);

        // 3.1 the index must probably be changed because of inverseLayerOrder
        // TODO Check
        if (me.inverseLayerOrder) {
            var totalInGroup = group.getLayers().getLength();
            layerIdx = totalInGroup - layerIdx - 1;
        }

        // 4. find the node that represents the group
        var parentNode;
        if (group === me.getLayerGroup()) {
            parentNode = me.getRootNode();
        } else {
            parentNode = me.getRootNode().findChildBy(function(candidate){
                return candidate.getOlLayer() === group;
            }, me, true);
        }
        if (!parentNode) {
            return;
        }

        // 5. insert a new layer node at the specified index to that node
        var layerNode = Ext.create('GeoExt.data.model.LayerTreeNode',
            layerOrGroup
        );
        layerNode.set('text', layerOrGroup.get(me.getTextProperty()));
        layerNode.commit();

        parentNode.insertChild(layerIdx, layerNode);

        if (layerOrGroup instanceof ol.layer.Group) {
            layerOrGroup.getLayers().forEach(me.addLayerNode, me);
        }
    },

    bindGroupLayerCollectionEvents: function(layerOrGroup) {
        var me = this;
        if (layerOrGroup instanceof ol.layer.Group) {
            var collection = layerOrGroup.getLayers();
            collection.on('remove', me.onLayerCollectionRemove, me);
            collection.on('add', me.onLayerCollectionAdd, me);
            collection.forEach(me.bindGroupLayerCollectionEvents, me);
        }
    },

    unbindGroupLayerCollectionEvents: function(layerOrGroup) {
        var me = this;
        if (layerOrGroup instanceof ol.layer.Group) {
            var collection = layerOrGroup.getLayers();
            collection.un('remove', me.onLayerCollectionRemove, me);
            collection.un('add', me.onLayerCollectionAdd, me);
            collection.forEach(me.unbindGroupLayerCollectionEvents, me);
        }
    },

    /**
     *  Remove children from rootNode and read the layerGroup-collection.
     *
     *  @private
     */
    onLayerCollectionAdd: function(evt){
        var me = this;
        if (me.collectionEventsSuspended) {
            return;
        }
        var layerOrGroup = evt.element;
        me.addLayerNode(layerOrGroup);
        me.bindGroupLayerCollectionEvents(layerOrGroup);
    },


    onLayerCollectionRemove: function(evt){
        var me = this;
        if (me.collectionEventsSuspended) {
            return;
        }
        var layerOrGroup = evt.element;
        // 1. find the node that existed for that layer
        var node = me.getRootNode().findChildBy(function(candidate) {
            return candidate.getOlLayer() === layerOrGroup;
        });
        if (!node) {
            return;
        }
        // 2. if grouplayer: cascade down and remove any possible listeners?
        if (layerOrGroup instanceof ol.layer.Group) {
            me.unbindGroupLayerCollectionEvents(layerOrGroup);
        }
        // 3. find the parent
        var parent = node.parentNode;
        // 4. remove the node from the parent
        parent.removeChild(node);
    },

    /**
     * Allows for temporarily unlistening to change events on the underlying
     * OpenLayers collections. Use #resumeCollectionEvents to start listening
     * again.
     */
    suspendCollectionEvents: function(){
        this.collectionEventsSuspended = true;
    },

    /**
     * Undoes the effect of #suspendCollectionEvents; so that the store is now
     * listening to change events on the underlying OpenLayers collections.
     * again.
     */
    resumeCollectionEvents: function(){
        this.collectionEventsSuspended = false;
    }
});
